// -*-c++-*-

/***************************************************************************
                                resultsaver.hpp
                    Abstract Base Class for saving results
                             -------------------
    begin                : 08-MAY-2003
    copyright            : (C) 2003 by The RoboCup Soccer Simulator 
                           Maintenance Group.
    email                : sserver-admin@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU LGPL as published by the Free Software  *
 *   Foundation; either version 2 of the License, or (at your option) any  *
 *   later version.                                                        *
 *                                                                         *
 ***************************************************************************/


#ifndef RCSSRESULTSAVER_HPP
#define RCSSRESULTSAVER_HPP


#include <string>
#include <rcssbase/factory.h>
#include <rcssbase/libloader.h>
#include <rcssbase/lib_shared_ptr.h>
#include <rcssbase/conf/parser.hpp>
#include <rcssbase/conf/builder.hpp>

struct tm;



namespace rcss
{
    class ResultSaver
        : public conf::Builder
    {
    public:
        typedef lib_shared_ptr< ResultSaver > Ptr;
        typedef Ptr(*Creator)( int, const char**, const std::string& module_name );    
        typedef ObjFact< Creator > Factory;

        enum team_id { TEAM_LEFT = 0, TEAM_RIGHT = 1 };

        ResultSaver( int, const char**, const std::string& module_name )
            : conf::Builder( module_name ),
              m_conf_parser( *this )
        {}

        virtual
        ~ResultSaver()
        {}

        bool
        enabled() const
        { return doEnabled(); }

        void
        saveStart()
        { doSaveStart(); }

        void
        saveTime( const tm& time )
        { doSaveTime( time ); }

        void
        saveTeamName( team_id team, const std::string& name )
        { doSaveTeamName( team, name ); }

        void
        saveCoachName( team_id team, const std::string& name )
        { doSaveCoachName( team, name ); }

        void
        saveScore( team_id team, unsigned int score )
        { doSaveScore( team, score ); }

        void
        savePenTaken( team_id team, unsigned int taken )
        { doSavePenTaken( team, taken ); }

        void
        savePenScored( team_id team, unsigned int scored )
        { doSavePenScored( team, scored ); }

        void
        saveCoinTossWinner( team_id team )
        { doSaveCoinTossWinner( team ); }

        bool
        saveComplete()
        { return doSaveComplete(); }

    protected:
        conf::Parser m_conf_parser;

    private:

        virtual
        bool
        doEnabled() const
        { return false; }

        virtual
        void
        doSaveStart()
        {}

        virtual
        void
        doSaveTime( const tm& )
        {}

        virtual
        void
        doSaveTeamName( team_id, const std::string& )
        {}

        virtual
        void
        doSaveCoachName( team_id, const std::string& )
        {}

        virtual
        void
        doSaveScore( team_id, unsigned int )
        {}

        virtual
        void
        doSavePenTaken( team_id, unsigned int )
        {}

        virtual
        void
        doSavePenScored( team_id, unsigned int )
        {}

        virtual
        void
        doSaveCoinTossWinner( team_id )
        {}

        virtual
        bool
        doSaveComplete()
        { return true; }
    };
}


#endif
